package be.iminds.ilabt.jfed.highlevel.util;

import be.iminds.ilabt.jfed.gui_preferences.HLPreferenceKey;
import be.iminds.ilabt.jfed.gui_preferences.JFedHLPreferences;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedConnection;
import be.iminds.ilabt.jfed.lowlevel.ssh_key_info.PuTTYFilesKeyInfo;
import be.iminds.ilabt.jfed.lowlevel.ssh_key_info.SshFilesKeyInfo;
import be.iminds.ilabt.jfed.lowlevel.ssh_key_info.SshKeyInfoFactory;
import be.iminds.ilabt.jfed.lowlevel.user.GeniUser;
import be.iminds.ilabt.jfed.lowlevel.user.GeniUserProvider;
import be.iminds.ilabt.jfed.preferences.CorePreferenceKey;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.ssh_terminal_tool.known_hosts.OpenSshKnownHostsFile;
import be.iminds.ilabt.jfed.ssh_terminal_tool.unix.SshAgentHelper;
import be.iminds.ilabt.jfed.util.common.Pair;
import be.iminds.ilabt.jfed.util.tmp_file_helpers.PermanentFile;
import be.iminds.ilabt.jfed.util.tmp_file_helpers.TmpFile;
import com.google.common.net.InetAddresses;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.ArrayList;
import java.util.List;

/**
 * SshCommandBuilder
 */
public class SshCommandBuilder {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(SshCommandBuilder.class);

    @Nonnull private final JFedHLPreferences jFedPreferences;
    @Nonnull private final OpenSshKnownHostsFile openSshKnownHostsFile;

    private String targetServerUserName;
    private String hostName;
    private int port = -1;
    private List<PuTTYFilesKeyInfo> sshKeyInfos;
    private JFedConnection.ProxyInfo proxyInfo;
    private boolean useSshAgent;
    private boolean forcePseudoTerminal = false;

    SshCommandBuilder(@Nonnull GeniUserProvider geniUserProvider,
                      @Nonnull JFedHLPreferences jFedPreferences,
                      @Nonnull OpenSshKnownHostsFile openSshKnownHostsFile) {
        this.jFedPreferences = jFedPreferences;
        this.openSshKnownHostsFile = openSshKnownHostsFile;

        List<PuTTYFilesKeyInfo> sshKeyInfos = new ArrayList<>();
        PuTTYFilesKeyInfo customSshKeyInfo = jFedPreferences.getCustomSshKeyInfo();
        if (customSshKeyInfo != null) {
            LOG.info("ExternalSshTerminal: The preferences are overriding the SSH key used (" + customSshKeyInfo.getClass().getName() + ").");
            sshKeyInfos.add(customSshKeyInfo);
        }
        if (jFedPreferences.getBoolean(CorePreferenceKey.PREF_SSH_USE_LOGIN_KEY, false) || sshKeyInfos.isEmpty()) {
            LOG.info("ExternalSshTerminal: The user's SSH key will be used.");
            sshKeyInfos.add(SshKeyInfoFactory.createGeniUserSshKeyInfo(geniUserProvider.getLoggedInGeniUser()));
        }
        this.sshKeyInfos = sshKeyInfos;
    }

    /**
     * @param useDummyValuesForKeysIfNeeded if a key is not directly available, and would require a tmp file. use a placeholder instead.
     *                                      Off course, this means the command becomes useless, except for showing to a user.
     * @return the ssh command and a list of tmp files needed by it.
     */
    public Pair<String, List<TmpFile>> build(boolean useDummyValuesForKeysIfNeeded) {
        if (!hasNeededData()) {
            throw new IllegalStateException("Cannot build: not all needed setters were called");
        }

        String resSshCommand;
        List<TmpFile> tmpFiles = new ArrayList<>();

        if (useSshAgent) {
            for (SshFilesKeyInfo sshKeyInfo : sshKeyInfos) {
                try {
                    SshAgentHelper.registerKey(sshKeyInfo);
                } catch (Exception e) {
                    LOG.warn("ssh-add failed: sshKeyInfo.hasUnencryptedPrivateKey=" + sshKeyInfo.hasUnencryptedPrivateKey(), e);
                }
            }
        }

        if (proxyInfo == null) {
            resSshCommand = jFedPreferences.getString(HLPreferenceKey.PREF_COMMAND_SSH);
            if (resSshCommand == null) {
                //we're probably on windows. Just use the default.
                resSshCommand = JFedHLPreferences.DEFAULT_UNIX_SSH_COMMAND;
//                throw new RuntimeException("No SSH command specified in jFed config. Cannot launch terminal.");
            }
        } else {
            assert proxyInfo instanceof JFedConnection.SshProxyInfo;

            JFedConnection.SshProxyInfo sshProxyInfo = (JFedConnection.SshProxyInfo) proxyInfo;

            //store proxyOpenSshKey
            if (sshProxyInfo.getHostKey() != null)
                openSshKnownHostsFile.addKnownHost(sshProxyInfo.getHostname(), sshProxyInfo.getHostKey());

            resSshCommand = jFedPreferences.getString(HLPreferenceKey.PREF_COMMAND_SSH_PROXY);
            if (resSshCommand == null) {
                //we're probably on windows. Just use the default.
                resSshCommand = JFedHLPreferences.DEFAULT_UNIX_PROXY_SSH_COMMAND_FOR_NEW_OPENSSH;

//                throw new RuntimeException("No SSH command specified in jFed config. Cannot launch terminal.");
            }

            resSshCommand = resSshCommand.replace("$P", sshProxyInfo.getPort() + "");
            resSshCommand = resSshCommand.replace("$H", sshProxyInfo.getHostname());
            resSshCommand = resSshCommand.replace("$U", sshProxyInfo.getUsername());

            if (sshProxyInfo.getSshKeyInfo() != null) {
                assert sshProxyInfo.getSshKeyInfo() instanceof SshFilesKeyInfo :
                        "sshProxyInfo.getSshKeyInfo() -> "+sshProxyInfo.getSshKeyInfo().getClass().getName();
                SshFilesKeyInfo sshFilesKeyInfo = (SshFilesKeyInfo)sshProxyInfo.getSshKeyInfo();
                if (useDummyValuesForKeysIfNeeded) {
                    TmpFile tmpFile = sshFilesKeyInfo.getPrivateKeyFile();
                    if (tmpFile instanceof PermanentFile) {
                        resSshCommand =
                                resSshCommand.replace("$I", tmpFile.getFilename());
                    } else {
                        resSshCommand =
                                resSshCommand.replace("$I", "<PEM file>");
                    }
                    //do not add to tmpFiles -> we don't want any file to be stored
                } else {
                    if (useSshAgent) {
                        TmpFile tmpFile = sshFilesKeyInfo.getPrivateKeyFile();
                        tmpFiles.add(tmpFile);
                        resSshCommand =
                                resSshCommand.replace("$I", tmpFile.getFilename()); //if using ssh-agent, just use encrypted file

                        try {
                            SshAgentHelper.registerKey(sshFilesKeyInfo);
                        } catch (SshAgentHelper.SshAgentHelperException e) {
                            LOG.error("Problem registering user SSH keys when starting terminal. Login might fail.", e);
                        }
                    } else {
                        TmpFile tmpFile = sshFilesKeyInfo.getUnencryptedPrivateKeyFile();
                        tmpFiles.add(tmpFile);
                        resSshCommand =
                                resSshCommand.replace("$I", tmpFile.getFilename());
                    }
                }
            }

            //workaround for IPv6 literal
            boolean isIpv6hostName = hostName.contains(":") &&
                    InetAddresses.isInetAddress(hostName);
            if (isIpv6hostName && resSshCommand.contains(" -W %h:%p")) {
                resSshCommand = resSshCommand.replace(" -W %h:%p", " -W [%h]:%p");
            }
        }

        if (forcePseudoTerminal) {
            resSshCommand = resSshCommand.replaceFirst("^ssh ", "ssh -tt ");
        }

        resSshCommand = resSshCommand.replace("$p", port + "");
        resSshCommand = resSshCommand.replace("$h", hostName);
        resSshCommand = resSshCommand.replace("$u", targetServerUserName);

        //OpenSSH option "-i <key>" can be specified multiple times, so we will do that.
        String iReplacement = null;
        boolean isDefaultIOption = resSshCommand.contains("-i $i");
        boolean isQuotedDefaultIOption = resSshCommand.contains("-i '$i'");
        boolean isDoubleQuotedDefaultIOption = resSshCommand.contains("-i \"$i\"");
        if (!(isDefaultIOption || isQuotedDefaultIOption || isDoubleQuotedDefaultIOption)) {
            LOG.warn("isDefaultIOption=false for sshCommand=\""+resSshCommand+"\"");
        }
        for (SshFilesKeyInfo sshKeyInfo : sshKeyInfos) {
            LOG.debug("launchTerminal sshKeyInfo class: " + sshKeyInfo.getClass().getName()+" isDefaultIOption="+isDefaultIOption+" iReplacement=\""+iReplacement+"\"");
            TmpFile privKeyFile = sshKeyInfo.getPrivateKeyFile();
            String privKeyFilename;
            if (useDummyValuesForKeysIfNeeded) {
                if (privKeyFile instanceof PermanentFile) {
                    privKeyFilename = privKeyFile.getFilename();
                } else {
                    privKeyFilename = "<PEM file>";
                }
                //do not add to tmpFiles -> we don't want any file to be stored
            } else {
                privKeyFilename = privKeyFile.getFilename();
                tmpFiles.add(privKeyFile);
            }
            if (privKeyFilename != null) {
                if (iReplacement == null) {
                    iReplacement = privKeyFilename;
                } else {
                    if (isDefaultIOption || isQuotedDefaultIOption || isDoubleQuotedDefaultIOption) {
                        if (isDefaultIOption)
                            iReplacement = iReplacement + " -i " + privKeyFilename;
                        if (isQuotedDefaultIOption)
                            iReplacement = iReplacement + "' -i '" + privKeyFilename;
                        if (isDoubleQuotedDefaultIOption)
                            iReplacement = iReplacement + "\" -i \"" + privKeyFilename;
                    } else {
                        LOG.warn("sshCommand used is not using default openSSH -i option -> Will only use first private key to login!");
                    }
                }
                //WAS:  sshCommand = sshCommand.replace("$i", privKeyFilename);
            } else {
                throw new RuntimeException("No private key file specified, cannot use SSH login. sshKeyInfo class: " + sshKeyInfo.getClass().getName());
            }

        }
        if (iReplacement != null) {
            LOG.debug("Using final iReplacement=\""+iReplacement+"\"");
            resSshCommand = resSshCommand.replace("$i", iReplacement);
        } else {
            throw new RuntimeException("Could not replace $i in sshCommand");
        }
        return new Pair<>(resSshCommand, tmpFiles);
    }

    public void setTargetServerUserName(@Nonnull String targetServerUserName) {
        this.targetServerUserName = targetServerUserName;
    }

    public void setHostName(@Nonnull String hostName) {
        this.hostName = hostName;
    }

    public void setPort(int port) {
        this.port = port;
    }

    /**
     * Good defaults are automatically set. But they can be overridden
     * @param sshKeyInfos
     */
    public void overrideSshKeyInfos(@Nonnull List<PuTTYFilesKeyInfo> sshKeyInfos) {
        this.sshKeyInfos = sshKeyInfos;
    }

    public void setProxyInfo(@Nullable JFedConnection.ProxyInfo proxyInfo) {
        this.proxyInfo = proxyInfo;
    }

    public void setUseSshAgent(boolean useSshAgent) {
        this.useSshAgent = useSshAgent;
    }

    public boolean hasNeededData() {
        return targetServerUserName != null && hostName != null && port > 0;
    }

    public void setForcePseudoTerminal(boolean forcePseudoTerminal) {
        this.forcePseudoTerminal = forcePseudoTerminal;
    }



    public String getTargetServerUserName() {
        return targetServerUserName;
    }

    public String getHostName() {
        return hostName;
    }

    public int getPort() {
        return port;
    }

    public List<PuTTYFilesKeyInfo> getSshKeyInfos() {
        return sshKeyInfos;
    }

    public JFedConnection.ProxyInfo getProxyInfo() {
        return proxyInfo;
    }

    public boolean isUseSshAgent() {
        return useSshAgent;
    }

    public boolean isForcePseudoTerminal() {
        return forcePseudoTerminal;
    }
}
